# Julia Form generator ********************************
# FormGen 0.2.1 30 JApril 2023
# BSD Zero Clause License free to use but no warranties
# El Condor - Condor Informatique - Turin
# *****************************************************
module formgen
using Dates
using Gtk, Gtk.ShortNames
using GTK3_jll 
const libgtk = libgtk3
const libgdk = libgdk3
const GTK_ALIGN_START = 1     # Snap to left or top side, leaving space on right or bottom
const GTK_ALIGN_BASELINE = 4  # Align the widget according to the baseline
macro defaultDateFormat() return "Y-m-d" end
function addControl(name,c)
        if !haskey(wdgControls,name)
                wdgControls[name] = [c]
        else
                a = wdgControls[name]
                push!(a,c)
                wdgControls[name] = a
        end
end
function ComposedWidget(wName,cBack,after=false)
        gridNV = GtkGrid()
        btnNext = GtkButton(">")
        btnPrevious = GtkButton("<")
        ibox = GtkGrid()
        set_gtk_property!(btnNext,:name,"fg_lButton")
        set_gtk_property!(btnPrevious,:name,"fg_lButton")
        gridNV[1,1] = GtkLabel("")
        ibox[1,1] = btnPrevious
        ibox[2,1] = btnNext
        gridNV[3,1] = GtkLabel("")
        if cBack != ""
                signal_connect(btnNext, "button-press-event") do widget, event
                        eval(Meta.parse("Main.$cBack(\"$wName\",\"Next\")"))
                end
                signal_connect(btnPrevious, "button-press-event") do widget, event
                        eval(Meta.parse("Main.$cBack(\"$wName\",\"Previous\")"))
                end
        end
        set_gtk_property!(gridNV, :column_homogeneous, true)
        if after return ibox
        else
                gridNV[2,1] = ibox
                return gridNV 
        end
end
function createOptions(name,items)
        cmb = wdgHandles[name]
        if haskey(wdgEventsId,name) signal_handler_block(cmb,wdgEventsId[name]) end
        empty!(cmb)
	wdgCombos[name], wdgCombosIndex[name] = splitKeyValue(items)
        for value in wdgCombosIndex[name]
                push!(cmb,value)
        end
        if haskey(wdgEventsId,name) signal_handler_unblock(cmb,wdgEventsId[name]) end
end
function checkData(endFunction = "")
        isOk = true
        signals = ""
        if endFunction != "" isOk = eval(Meta.parse("Main.$endFunction(\"End\")")) end
        Operators = ["==","!=","<","<=",">",">="]
        for key in keys(wdgControls)            
                controls = wdgControls[key]
                lblHandle = wdgLabelHandles[key]
                value = valueOf(key,true)
                if isWindows
                        lblTxt = get_gtk_property(lblHandle,:label,String)
                        if startswith(lblTxt,errSignal)
                                lblTxt = replace(lblTxt,errSignal=>"")
                                set_gtk_property!(lblHandle,:label,lblTxt) 
                        end
                else
                        fg_set_gtk_style(wdgLabelHandles[key], "#fg_label {color:$labelColor}",1)
                end
                if value == "" && key ∉ wdgRequired continue end
                for i in eachindex(controls)
                        Operator, Operand, Signal = controls[i]
                        if value == "" && Operand != ".+" continue end     # no check if empty
                        nValue = tryparse(Float64, value)
                        if haskey(wdgHandles,Operand) Operand = valueOf(Operand) end
                        nOperand = tryparse(Float64, Operand)
                        if Operator in Operators
                                if nValue !== nothing && nOperand !== nothing
                                        if !eval(Meta.parse("$value $Operator $Operand"))
                                                signals = "$signals\n$key: $Signal"
                                                if isWindows set_gtk_property!(lblHandle,:label,errSignal * lblTxt)
                                                else fg_set_gtk_style(wdgLabelHandles[key], "#fg_label {color:red}",1)
                                                end
                                                isOk = false
                                        end
                                else    # alfanumeric compare                                       
                                        if !eval(Meta.parse("\"$value\" $Operator \"" * valueOf(Operand) * "\""))
                                                signals = "$signals\n$key: $Signal"
                                                if isWindows set_gtk_property!(lblHandle,:label,errSignal * lblTxt)
                                                else fg_set_gtk_style(wdgLabelHandles[key], "#fg_label {color:red}",1)
                                                end
                                                isOk = false
                                        end
                                end
                        elseif lowercase(Operator) == "is"
                                if !occursin(Regex(Operand),value)
                                        signals = "$signals\n$key: $Signal"
                                        isOk = false
                                        if isWindows set_gtk_property!(lblHandle,:label,errSignal * lblTxt)
                                        else fg_set_gtk_style(wdgLabelHandles[key], "#fg_label {color:red}",1)
                                        end
                        end
                        end
                end
        end
        if signals != "" info_dialog(signals) end
        return isOk
end
function createAfter(fName)     # afther name
        wdg = wdgAfter[fName]
        name = wdg[2]
        if wdg[1] == "NV" return ComposedWidget(name,wdg[3],true)
        else
                lbl = wdg[3]              
                btn = createButton(name,lbl)
                functionCall = getParameter(wdg[1:end],"call")
                if functionCall != ""
                        push!(wdgEvents,["Event", name, "clicked", "call", functionCall, "link", fName])
                        wdgHandles[name] = btn
                else
                        signal_connect((widget, event) -> callback_ok(widget),btn, "button-press-event")

                end
                return btn
        end
end
function has(h,parm)  # returns true or false
        return in(parm,map(x->lowercase(x),h))
end
function getParameter(h,parm,default = "")  # returns a word of h that is after parm
        re = Regex("^" * parm * "\$")   # parm can be x|y
        for i = 1:lastindex(h)
                if occursin(re,lowercase(h[i]))
                        if i < length(h) return h[i+1] end
                end
        end
        return default
end
function fg_set_CSS(css::String)        # set CSS at screen level
        provider = convert(GObject,ccall((:gtk_css_provider_new, libgtk), Ptr{GObject},()))
        display = convert(GObject,ccall((:gdk_display_get_default,libgdk), Ptr{GObject}, ()))
        ccall((:gdk_display_flush,libgdk),Cvoid,(Ptr{GObject},),display)
        screen = convert(GObject,ccall((:gdk_display_get_default_screen,libgdk),Ptr{GObject},(Ptr{GObject},),display))
        ccall((:gtk_css_provider_load_from_data,libgtk),Bool,(Ptr{GObject}, Ptr{UInt8}, Clong, Ptr{C_NULL}),
                provider, css,-1, Ptr{Ptr{Cchar}}())
        ccall((:gtk_style_context_add_provider_for_screen,libgtk),Cvoid,(Ptr{GObject},Ptr{GObject}),screen,provider)
        return provider
end
function fg_set_gtk_property(widget::String, name::Core.Symbol, value::Any)	# called with name
	set_gtk_property!(wdgHandles[widget],name,value)
	return nothing
end
function fg_set_gtk_property(widget::Gtk.GtkWidget, name::Core.Symbol, value::Any) # called with handle
	set_gtk_property!(widget,name,value)
	return nothing
end
function fg_reset_CSS(provider)
        display = convert(GObject,ccall((:gdk_display_get_default,libgdk), Ptr{GObject}, ()))
        screen = convert(GObject,ccall((:gdk_display_get_default_screen,libgdk),Ptr{GObject},(Ptr{GObject},),display))
        ccall((:gtk_style_context_remove_provider_for_screen,libgtk),Cvoid,(Ptr{GObject},Ptr{GObject}),screen,provider)
end
# ***************** set style ***********************
function fg_set_gtk_style(wdgName::String, style::String, value::Int=600)	# called with name
	return fg_set_gtk_style(wdgHandles[wdgName],style,value)
end
function fg_set_gtk_style(widget::Vector{GObject}, style::String, value::Int=600) # called with text and buffer handle
	return fg_set_gtk_style(widget[3],style,value)
end
function fg_set_gtk_style(widget::Gtk.GtkWidget, style::String, value::Int=600)   # called with handle # thank to MeteorBizone
        sc = Gtk.GAccessor.style_context(widget)
        pr = Gtk.CssProviderLeaf(data=style)
        push!(sc, Gtk.StyleProvider(pr), value)
        return nothing
end
function fg_handle_2_name(widget)      
        if @isdefined wdgHandles
                w = [k for (k,v) in wdgHandles if v == widget]
                if length(w) > 0 return w[1] end
        end
        return ""
end
function fg_create_textview(;width=0,height=0)
        textBuf = TextBuffer()
        tArea = TextView(textBuf)
        sWind = convert(GObject, ccall((:gtk_scrolled_window_new,libgtk),Ptr{GObject},(Ptr{UInt8},Ptr{UInt8}),Ptr{UInt8}(0),Ptr{UInt8}(0)))
        set_gtk_property!(sWind,:name,"fg_textArea")
        if width + height > 0               
                style = width > 0 ? "min-width:$width" * "px;" : ""
                if height > 0 style = "$style" * "min-height:$height" * "px;" end
        end
        ccall((:gtk_container_add,libgtk),Cvoid,(Ptr{GObject},Ptr{GObject}),sWind,tArea)
        fg_set_gtk_style(sWind,"* {$style}",600)
        return [textBuf,tArea,sWind]
end
function fg_menu_call(fnz)
	if occursin(r"close|exit",lowercase(fnz)) destroy(fg_win)
	else info_dialog(fnz)
	end
end
function fg_create_menu(menuList)	
        mn = Dict()
        mnOrder = []    # for preserve order
        for i in eachindex(menuList)
                field = menuList[i]
                key = field[1]
                if !haskey(mn, key) 
                        mn[key] = [field]
                        push!(mnOrder,key)
                else
                        push!(mn[key],field)
                end
        end
        menuBar = GtkMenuBar()
        for i in eachindex(mnOrder)
                menuItems = mn[mnOrder[i]]
                menuItem = GtkMenuItem(menuItems[1][1]) # also father creates a menu item
                fg_set_gtk_property(menuItem,:name,"fg_menuButton") 
                menu = GtkMenu(menuItem)
                for j in eachindex(menuItems)
                        mnItem = GtkMenuItem(menuItems[j][2])
                        push!(menu,mnItem)
                        fnz = menuItems[j][3]
                        prm = menuItems[j][2]
                        if isdefined(Main,Symbol(fnz)) && getproperty(Main,Symbol(fnz)) isa Function             
                                id = signal_connect(mnItem, "activate") do widget           
                                        eval(Meta.parse("Main.$fnz(\"$prm\")"))
                                end
                        else                                                               
	                        signal_connect(mnItem, "activate") do widget
	                                fg_menu_call(fnz)
	                        end
                        end
                end
            push!(menuBar, menuItem)
        end
        return menuBar
end
function fg_DateObject(handleCal)       # set date object
        Y = Ref{Cuint}(0)       # CHAPTER 27. CALLING C AND FORTRAN CODE
        M = Ref{Cuint}(0)       # Passing Pointers for Modifying Inputs
        D = Ref{Cuint}(0)
        ccall((:gtk_calendar_get_date,libgtk), Cvoid,(Ptr{GObject},Ref{Cuint},Ref{Cuint},Ref{Cuint}),handleCal,Y,M,D)      
        return DateTime(Y[],M[]+1,D[])
end
function fg_DateSelected(handleCal,name)
        global fg_winDate
        dt = fg_DateObject(handleCal)
        setValue(name,Dates.format(dt,@defaultDateFormat))       # the request format is set in setValue
        @idle_add destroy(fg_winDate)
        nothing
end
function fg_create_calendar(fnz,value = "",format = @defaultDateFormat;name="")
        function fg_MonthYear(x,id)
                signal_handler_block(x,id)
                @idle_add signal_handler_unblock(x,id)
                nothing
        end
        c = convert(GObject, ccall((:gtk_calendar_new,libgtk), Ptr{GObject}, ()))
        if fnz == "fg_DateSelected"
                dt = wdgDate[name]
        else
                if value != ""
                        if lowercase(value) == "tomorrow" value = Dates.format(Dates.now()+Dates.Day(1),"Ymd") end
                        value = replace(value,r"^(\d{4})\D?(\d\d?)\D?(\d\d?)$|^(\d{4})(\d\d)(\d\d)$" => s"\g<1>-\g<2>-\g<3>")
                        dt = tryparse(Date,value,dateformat"Y-m-d")
                else 
                        dt = Dates.now() 
                end
                if dt !== nothing        
                        Y = Cuint(Dates.year(dt))       # CHAPTER 27. CALLING C AND FORTRAN CODE
                        M = Cuint(Dates.month(dt)-1)
                        D = Cuint(Dates.day(dt))
                        ccall((:gtk_calendar_select_day,libgtk), Cvoid,(Ptr{GObject},Cuint,),c,D)
                        ccall((:gtk_calendar_select_month,libgtk), Cvoid,(Ptr{GObject},Cuint,Cuint),c,M,Y)
                end
        end
        if fnz == fg_DateSelected
                id = signal_connect(w->fg_DateSelected(w, name), c, "day-selected")
        else
                id = signal_connect(c, "day-selected") do widget
                        dt = fg_DateObject(widget)
                        if format != "" 
                                eval(Meta.parse("Main.$fnz(\"" * Dates.format(dt,format) *"\")"))
                                 
                        else
                                eval(Meta.parse("Main.$fnz(\"" * Dates.format(dt,"Y-m-d") *"\")"))
                        end                     
                end
        end              
        signal_connect(w->fg_MonthYear(c,id), c, "next-month")
        signal_connect(w->fg_MonthYear(c,id), c, "next-year")
        signal_connect(w->fg_MonthYear(c,id), c, "prev-month")
        signal_connect(w->fg_MonthYear(c,id), c, "prev-year")
        return c
end
function fg_Date(btn,dateName,title)
        c = fg_create_calendar(fg_DateSelected,Dates.format(wdgDate[dateName],@defaultDateFormat);name=dateName)
        global fg_winDate = GtkWindow("$title")
        push!(fg_winDate,c)
        showall(fg_winDate)
end
function fg_File(btn,fileName)
        field = wdgFields[fileName]
        if field[1] == "FOLDER"
                answer = open_dialog("Select Dataset Folder", action=GtkFileChooserAction.SELECT_FOLDER)
        else
                extra = field[4:length(field)]
                title = getParameter(extra,"title","Pick a file")
                if !has(field,"filter")
                        answer = open_dialog(title)
                else 
                        filter = getParameter(extra,"filter")
                        answer = open_dialog("$title", GtkNullContainer(), (replace(filter,r",\s*" => s", "),), )
                end
        end
        if answer != "" setValue(fileName,answer) end
end
function fieldClean(f)  # arrange label and extra
        push!(f,"","","","")
        if f[3] == ""   # label creation
                f[3] = replace(f[2],r"_[a-zA-Z]" => x -> " " * lowercase(SubString(x,2)))
                f[3] = replace(f[3],r"[a-z][A-Z]" => x -> SubString(x,1,1) * " " * lowercase(SubString(x,2)))
                f[3] = uppercase(SubString(f[3],1,1)) * SubString(f[3],2)
        end
end
function createButton(name,caption)
        btn = GtkButton(caption)
        if length(caption)== 1 set_gtk_property!(btn, :name, "fg_lButton") 
        else set_gtk_property!(btn, :name, "fg_button")
        end
        return btn
end
function getDateNames(type)
        t = Date(2018, 1, 1)    # because is sunday
        typeParm = lowercase(type)
        r = []
        if typeParm == "%days" || typeParm == "%#days"    
                for i = 1:1:7
                        push!(r, first(type,2) == "%#" ? "$i:"*Dates.dayname(t) : Dates.dayname(t))
                        t = t + Dates.Day(1)
                end
        elseif typeParm == "%months" || typeParm == "%#months"
                for i = 1:1:12
                        push!(r,first(type,2) == "%#" ? "$i:"*Dates.monthname(t) : Dates.monthname(t))
                        t = t + Dates.Month(1)
                end
        else
                return type # no need to convert
        end
        return join(r,"|")
end
function setAlignment(f,default)
        alignment = alignmentConst[default]
        for i = 1:lastindex(f)
                if haskey(alignmentConst, uppercase(f[i]))
                        alignment = alignmentConst[uppercase(f[i])]   
                end
        end
        return alignment
end
function setValues()
	for w in wdgDefaults
		key,value = w	
                setValue(key, value)             
        end
end
function hasDefault(name)
	for w in wdgDefaults
		key,v = w
		if key == name && v != "" return true end                    
        end
        return false
end
function hasDefault(name,value)
	for i = 1:length(wdgDefaults)
		key,v = wdgDefaults[i]
		if key == name 
			if v == "" wdgDefaults[i] = name => value end
			return
		end                    
        end
        append!(wdgDefaults,[name => value])
end
function setValue(key,value)
        type = get(wdgTypes, key, "")
        if type in textWidgets
                set_gtk_property!(wdgHandles[key], :text, value)
        elseif type == "CKB"
        	set_gtk_property!(wdgHandles[key],:active,false)	# reset               
                set_gtk_property!(wdgHandles[key],:active,uppercase(value) == "ON" ? true : false)
        elseif type == "CMB"
                if haskey(wdgEventsId,key) signal_handler_block(wdgHandles[key],wdgEventsId[key]) end
                indx = -1
                for (comboValue, comboKey) in wdgCombos[key]
                        if comboKey == value
                                indx = findfirst(z -> z == comboValue,wdgCombosIndex[key]) - 1
                                break
                        end
                end
                @idle_add set_gtk_property!(wdgHandles[key],:active,indx)
                if haskey(wdgEventsId,key) @idle_add signal_handler_unblock(wdgHandles[key],wdgEventsId[key]) end
        elseif type == "H"
                wdgHiddens[key] = value
        elseif type == "DATE"
                if lowercase(value) == "tomorrow" value = Dates.format(Dates.now()+Dates.Day(1),"Y-m-d") end
                value = replace(value,r"^(\d{4})\D?(\d\d?)\D?(\d\d?)$|^(\d{4})(\d\d)(\d\d)$" => s"\g<1>-\g<2>-\g<3>")
                dt = tryparse(Date,value,dateformat"Y-m-d")
                if dt === nothing 
                        println("Default date $value incorrect")
                else
                        field = wdgFields[key]
                        dtFormat = getParameter(field[3:length(field)],"format",@defaultDateFormat)
                        set_gtk_property!(wdgHandles[key], :text,Dates.format(dt,dtFormat))
                        wdgDate[key] = dt
                end
        elseif type == "RDB"
                hRdb = get(rdbHandlers,key * value,rdbHandlers[key])
                set_gtk_property!(hRdb,:active,true)    # hidden rdb active
                for k in keys(wdgCombos[key])
                       if wdgCombos[key][k] == value 
                                value = k
                                break
                       end
                end
                set_gtk_property!(wdgHandles[key], :text, value)
        elseif type == "S"
                if tryparse(Float64, value) !== nothing
                        G_.value(wdgHandles[key],parse(Float64, value)) # G_ is Gtk.GAccessor
                end
        elseif type == "TV"
                set_gtk_property!(wdgHandles[key][1],:text,value)
        elseif type == "U" || type == "UN"
                GAccessor.text(wdgHandles[key],"$value")
                set_gtk_property!(wdgHandles[key],:margin,5)
        elseif type == "C"
                GAccessor.text(wdgHandles[key],"$value")            
        end
end
function splitKeyValue(s)	# indexes is populated by values
        items = Dict{String,String}()
        indexes = []
        for item in split(s,"|")
                item = split("$item:$item",":")
                items[item[2]] = item[1]
                push!(indexes,item[2])
        end
        return items,indexes
end
function valueOf(name,retString=false)
        type = wdgTypes[name]
        if type in textWidgets || type == "DATE"
                value = get_gtk_property(wdgHandles[name],:text,String)
                if retString return value end   # for check purpose
                if haskey(numberWidgets,type) && value != ""
                        return parse(Float64, value)
                end
                return value
        elseif type == "H" return wdgHiddens[name]
        elseif type == "CMB"
                idx = get_gtk_property(wdgHandles[name], :active, Int)
                if idx > -1
                        str = Gtk.bytestring(GAccessor.active_text(wdgHandles[name]))
                        return wdgCombos[name][str]
                end
                return ""
        elseif type == "CKL"
        	ck = 0
        	for value in values(wdgCombos[name])
        		if get_gtk_property(wdgHandles[value], :active, Bool) == true ck = ck + 1 end
        	end
        	return ck
        elseif type == "CKB"
                return get_gtk_property( wdgHandles[name],:active,String)
        elseif type == "RDB"
                str = get_gtk_property(wdgHandles[name],:text,String)
                return (str == "") ? "" : wdgCombos[name][str]
        elseif type == "S"
                return string(Gtk.GAccessor.value(wdgHandles[name]))
        elseif type == "TV"
                return get_gtk_property(wdgHandles[name][1], :text, String)
        elseif type in ["U","UN","C"]
                return get_gtk_property(wdgHandles[name],:label,String)
        end
        return ""
end
function callback_numbers(widget, event,nType, eventType)
        value = get_gtk_property(widget,:text,String)
        if eventType == "down"
                global textBefore = value
                return nothing
        end
        k = event.keyval
        if k > 47 && k < 59 return nothing end
        if k < 255
        	if nType == "PN" set_gtk_property!(widget, :text, textBefore)
                elseif !occursin(Regex("^[+-]\$"),value)
                	if occursin(Regex("[+-]\$"),value)
                		value = value[end] * value[1:end-1]
                		set_gtk_property!(widget, :text, value)
                	end
	        	if !occursin(Regex(numbersCheckRegExpr[nType]), value)
			        set_gtk_property!(widget, :text, textBefore)
			end
                end                            
        end
        return nothing
end
function callback_ok(widget)
        global onEndFunction       	
	if checkData(onEndFunction)
	       	data = Dict{String,Any}()
	        data["fg_button"] = fg_handle_2_name(widget)
	        for key in keys(wdgTypes)
	                data[key] = valueOf(key)
	       	end
	        callBack(data)
	        if window["reset"] setValues() end
	        if !window["static"] destroy(fg_win) end
	end
end
function callback_rdb(widgetptr::Ptr, fieldName)
        widget = convert(Button, widgetptr)  # pointer -> object
        if get_gtk_property(widget, :active, Bool)
                set_gtk_property!(wdgHandles[fieldName], :text, get_gtk_property(widget, :label, AbstractString))
        end
        return nothing
end
function formGen(frm,answer)
global onEndFunction = ""	# on end
global isWindows = Sys.iswindows()
global labelColor = "black"
global screen_provider
global callBack = answer
global window = Dict{String,Any}("title"=>"Formgen","static"=>false,"reset"=>false,"ground"=>"white")
regex = r"\"(.*?)\"[ ]*|'(.*?)'[ ]*|([^ ]+)[ ]*"
global row = 1
global wdg = []
global alignmentConst = Dict{String,Any}("CENTER" => 0.5,"LEFT" => 0,"RIGHT" => 1)
global wdgDefaults = []				# name => default
global wdgHandles = Dict{String,Any}()          # name => handle
global wdgLabelHandles = Dict{String,Any}()	# name => label handle
global wdgTypes = Dict{String,String}()         # name => type
global wdgCombos = Dict{String,Dict}()          # combo and radiobuttons items
global wdgCombosIndex = Dict{String,Any}()      # combo values index (for preserve order)
global wdgControls = Dict{String,Any}()         # field controls
global wdgFields  = Dict{String,Any}()          # fields key => widget
global rdbHandlers = Dict{String,Any}()         # key = fieldNameValue
global wdgStyles = Dict{String,String}()        # key = fieldName
global wdgHiddens = Dict{String,Any}()         	# key = hiddenValue
global textBefore = ""                          # text on key_press
global wdgRequired = []                         # required fields
global wdgAfter = Dict{String,Any}()            # after buttons (for now)
global wdgDate = Dict{String,Any}()             # date exposed
global wdgEventsId = Dict()                     # events id
global wdgEvents = []                           # events container
menu = []                                       # menu container
standardButtons = Dict{String,String}("Ok"=>"Ok","Cancel"=>"Cancel","Reset"=>"Reset")
#global numbersRegExpr = Dict{String,Any}("FN"=>r"^[+-]?\d*\.{0,1}\d*$","N"=>r"^[+-]?\d*$","PN"=>r"^\d+$")
global numbersCheckRegExpr = Dict{String,String}("FN"=>"^[+-]?[0-9]*[.]?[0-9]+\$","N"=>"^[+-]?[0-9]+\$","PN"=>"^[0-9]+\$")
reNotEmpty = ".+"
global textWidgets = ["T","P","N","FN","PN","F","FOLDER"]
global errSignal = "***"
global numberWidgets = Dict{String,String}("UN"=>"20","N"=>"22","PN"=>"24","FN"=>"20")    # for output data, value is space for reduce width
types = Dict{String,String}("COMMENT"=>"C","TEXT"=>"T","NUMBER"=>"N","SLIDER"=>"S","HIDDEN"=>"H",
        "NAVIGATOR"=>"NV","REQUIRED"=>"REQ","FILE"=>"F","PSW"=>"P","COMBO"=>"CMB","CHECK"=>"CKB")
CSSStyles = """
#fg_button {margin:5px;background-image: image(#ccc)}
#fg_button:hover {background-image: image(#eee)}
#fg_button:active {background-image: image(#ddd)}
#fg_lButton {font: 20px \"sans\";min-width:20px;margin-left:3px;background-image: image(#ccc)}
#fg_lButton:hover {background-image: image(#eee)}
#fg_lButton:active {background-image: image(#ddd)}
#fg_navigButtons {min-width:10px;margin-left:3px}
#fg_menuButton {font: 16px \"sans\"}
#fg_scale {min-width: 160px;}
#fg_lScale {min-width: 140px;}
"""
labelsStyle = "\n#fg_label {margin:5px;}"
labelsAlignment = alignmentConst["RIGHT"]
labelsAfter = ""
wdgx = split(frm,"\n";keepempty=false)
for (index, value) in enumerate(wdgx)
        f = collect(eachmatch(regex,"$value"))  # separe fields items
        field = []
        for i in f
                push!(field,i.captures[findfirst(x -> !isnothing(x),i.captures)])
        end
        field[1] = get(types,uppercase(field[1]),uppercase(field[1]))      
        type = field[1]
        name = field[2]
        extra = field[4:end]
        if SubString(type,1,1) != "#"
                after = getParameter(extra,"after")      # check if has an after parameter
                if after != ""
                        wdgAfter[after] = field
                        if type == "NV" continue end
                end
                width = getParameter(extra,"width")      # check if has width
                if width != ""
#                        println("Style $name 'min-width:$width"*"px'")
                        push!(wdgx,"Style $name 'min-width:$width"*"px'")
                end
                if type == "B"
                        fieldClean(field)
                        if haskey(standardButtons,name)
                                standardButtons[name] = field[3]    # caption changed
                        end                        
                elseif type == "CONTROL"
                        push!(field,"","","")
                        if lowercase(field[3]*field[4]) == "ismail"
                                field[4] = "^[a-zA-Z0-9._-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,4}\$"    
                        end
                        addControl(name,[field[3],field[4],field[5]])
                elseif type == "DEFAULT"
                        for i = 2:2:length(field)-1
                        	hasDefault(field[i],field[i+1])
                        end
                elseif type == "H"
		        wdgTypes[name] = "H"
		        append!(wdgDefaults,[name => field[3]])
		        wdgHiddens[name] = field[3]
                elseif type == "LABEL"
                        labelsAlignment = setAlignment(field[2:end],"RIGHT")
                        labelsAfter = getParameter(field[2:end],"after","")
                        labelColor = getParameter(field[2:end],"color","black")
                        if labelColor != "" 
                                labelsStyle = replace(labelsStyle,"}" => "color:$labelColor;}")
                        end
                elseif type == "MENU"
                        push!(menu,field[2:end])
                elseif type == "ON" || type == "EVENT"
                        push!(wdgEvents,field)
                elseif type == "REQ"
                         for i in collect(2:lastindex(field))
                                push!(wdgRequired,field[i])
                        end
                elseif type == "STYLE"
		        wdgStyles[field[2]] = field[3]
		elseif type == "WINDOW"
                        wExtra = field[3:end]
                        window["title"] = name
                        if has(wExtra,"static") window["static"] = true end
                        if has(wExtra,"reset") window["reset"] = true end
                        ground = getParameter(wExtra,"(ground|background)")
                        if ground != "" push!(wdgx,"Style fg_window 'background:$ground'") end
                        fnz = getParameter(wExtra,"start|onstart")
                        if fnz != "" push!(wdgEvents,["On", "fg_window", "Start", fnz]) end
                        fnz = getParameter(wExtra,"end|onend")
                        if fnz != "" push!(wdgEvents,["On", "fg_window","End", fnz]) end
                else
                        fieldClean(field)
                        if type == "DATE"
                                title = getParameter(field[3:length(field)],"title",field[3])
                                push!(wdgx,"B fg_$name  \u25a6 After $name Call 'fg_Date $title'")
                        elseif type == "F" || type == "FOLDER"
                                push!(wdgx,"B fg_$name  \u22EF After $name Call fg_File")
                        end              
                        push!(wdg,field)
                        wdgTypes[name] = field[1]
                        wdgFields[name] = field
                        if !hasDefault(name) && type != "CKL"
                        	default = getParameter(extra,"(value|default)")   
                        	append!(wdgDefaults,[name => default])
                        end
                end

        end
end
for i in eachindex(wdgRequired)
        addControl(wdgRequired[i],["is", reNotEmpty, "The field is required"])
end
#if window["ground"] != ""
#        CSSStyles = CSSStyles * "#fg_window {background:$(window["ground"])}"
#end
title = window["title"]
global fg_win = GtkWindow("$title")
wdgHandles["fg_window"] = fg_win
@idle_add screen_provider = fg_set_CSS(CSSStyles * labelsStyle)
set_gtk_property!(fg_win, :name,"fg_window")
g = GtkGrid()
# create form *************************************************
if length(menu) > 0
        g[1,row] = fg_create_menu(menu)
        row = row + 1
end
oneCombo = false
for (field) in wdg
        # label creation
        rTime = Dates.now()
        type = field[1]
        name = field[2]
        extra = field[4:length(field)]
        supportGrid = GtkGrid()   # for numeric fields, cklist and after
        local label = GtkLabel(field[3] * labelsAfter)        # here for comment
        wdgLabelHandles[field[2]] = label       
        if type ∉ ["C","NV"]
                g[1,row] = label
                set_gtk_property!(label, :xalign, labelsAlignment)
                set_gtk_property!(label,:name,"fg_label")
        end
        if type in textWidgets || type == "DATE"
                if has(extra,"height")  # text view
                        height = parse(Int,getParameter(extra,"height","0"))
                	width = parse(Int,getParameter(extra,"width","150"))	# GTKEntry minimum-width is hardcoded 150px                        
                        handlers = fg_create_textview(height=height,width=width)
                        wdgHandles[name] = handlers  # buffer, textarea and scrollarea
#                        wdgHandles[name] = [handlers[1],handlers[2],handlers[3]]  # buffer and textarea
                                               
                        supportGrid[1,1] = handlers[3]  # the scroll view
                        if haskey(wdgAfter,field[2]) supportGrid[2,1] = createAfter(field[2]) end
                        g[2,row] = supportGrid
                        wdgTypes[name] = "TV"   # type TextView                      
                else
                        local txt = GtkEntry()
                        supportGrid[1,1] = txt
                        wdgHandles[name] = txt
                        hint = getParameter(extra,"hint")
                        if hint != ""
                                ccall((:gtk_entry_set_placeholder_text,libgtk),Cvoid,(Ptr{GObject},Ptr{Cchar}), txt,"$hint")
                        end
                        if type == "DATE"
                       		dtFormat = getParameter(field[3:end],"format",@defaultDateFormat)
                        	d = Dates.format(Dates.now(),dtFormat)                     
                        	set_gtk_property!(txt,:max_width_chars,length(d))
                        end
                        if type == "P" set_gtk_property!(txt,:visibility,false) end
                        fnz = getParameter(extra,"onactivate")
                        if fnz != "" push!(wdgEvents,["On",name, "activate", fnz]) end
                        if type in ["N","FN","PN"]
                        	fg_set_gtk_style(txt,"* {min-width:" * getParameter(extra,"width","150") * "px}") # GTKEntry minimum-width is hardcoded
                        	set_gtk_property!(txt,:xalign,1)        # right
                                supportGrid[2,1] = GtkLabel(repeat(" ",10))
                                signal_connect((widget, event) -> callback_numbers(widget, event,type,"down"),txt,"key-press-event")         
                                signal_connect((widget, event) -> callback_numbers(widget, event,type,"up"),txt,"key-release-event")
                                regExpr = get!(numbersCheckRegExpr,type,numbersCheckRegExpr["FN"])
                                addControl(name,["is", regExpr, "Not numeric field"])
                        end
                        if haskey(wdgAfter,name)
                        	supportGrid[2,1] = createAfter(name)
                	end
#                        if haskey(wdgAfter,name)
#                        	supportGrid[1,1] = txt
                        	g[2,row] = supportGrid
#                        else 
#                        	g[2,row] = txt 
#                        end                       
                end
        elseif type == "CMB"
                local cmb = GtkComboBoxText()
                supportGrid[1,1] = cmb
                set_gtk_property!(supportGrid, :column_homogeneous, true)
                wdgHandles[name] = cmb
                if haskey(wdgAfter,field[2]) supportGrid[2,1] = createAfter(field[2]) end
                createOptions(field[2],getDateNames(field[4]))
                g[2,row] = supportGrid
                if length(wdgTypes) == 1        # only one combo
                        oneCombo = true
                        wdgEventsId[field[2]] = signal_connect(cmb, "changed") do widget
                                data = Dict("fg_button"=>field[2],field[2] => valueOf(field[2]))
                                if window["reset"] setValues() end
                                if !window["static"] destroy(fg_win) end
                                callBack(data)
                        end
                end
                fnz = getParameter(extra,"onchange")
                if fnz != "" push!(wdgEvents,["On",name, "changed", fnz]) end

        elseif type == "C"      # comment
                alignment = setAlignment(field[4:length(field)],"LEFT")
                set_gtk_property!(label, :halign, GTK_ALIGN_BASELINE) # top left
                set_gtk_property!(label, :xalign, alignment)
                height = getParameter(extra,"height")
                if height != "" set_gtk_property!(label,:height_request,parse(Int32,height)) end
                GAccessor.line_wrap(label,true)
                if getParameter(extra,"color") != "" fg_set_gtk_style(label, "* {color:" * getParameter(extra,"color") * "}",1) end
                g[1:2,row] = label
                wdgHandles[name] = label               
                append!(wdgDefaults,[name => field[3]])
                set_gtk_property!(label,:name,"fg_comment")               
        elseif type == "NV"      # Navigator
                g[1:2,row] = ComposedWidget(name,field[3])
        elseif type == "U" || type == "UN"
                uLabel = GtkLabel("")
                supportGrid[1,1] = uLabel
                if type == "UN" 
                        supportGrid[2,1] = GtkLabel("")
                        set_gtk_property!(uLabel,:xalign,1)        # right
                end      # numeric field less wide
                g[2,row] = supportGrid
                wdgHandles[name] = uLabel
        elseif type == "CKB"
                ckb = GtkCheckButton(field[4])
                g[2,row] = ckb
                wdgHandles[name] = ckb
        elseif type == "CKL"	# vertical chek list
		i = 0
		default = getParameter(extra,"(value|default)")
		wdgCombos[name], wdgCombosIndex[name] = splitKeyValue(field[4])
		for value in wdgCombosIndex[name]
		        i = i + 1
		        supportGrid[1,i] = GtkCheckButton(value)		  
		        key = wdgCombos[name][value]
		        wdgHandles[key] = supportGrid[1,i]
		        hasDefault(key,default == key ? "on" : "")
		        wdgTypes[key] = "CKB"		       
		end         
                if haskey(wdgAfter,field[2]) 
                	afterPosition = trunc(Int,ceil(length(wdgCombosIndex[name])/2))
                	supportGrid[2,afterPosition] = createAfter(field[2]) 
                end
                g[2,row] = supportGrid     
        elseif type == "RDB"
                txt = GtkEntry()
                set_gtk_property!(txt, :text, "")       # contain rdb exposed value
                wdgHandles[name] = txt
                rdbGrid = GtkGrid()
                wdgCombos[name], wdgCombosIndex[name] = splitKeyValue(field[4])
                r = Vector{RadioButton}(undef, length(wdgCombosIndex[name])+1)
                r[1] =  RadioButton("")
                rdbHandlers[name] = r[1]
                vertical = has(extra,"vertical")
                for i = 2:length(wdgCombosIndex[name])+1
                        r[i] = RadioButton(r[i-1],wdgCombosIndex[name][i-1])
                        rdbHandlers[name * wdgCombos[name][wdgCombosIndex[name][i-1]]] = r[i]
                        if vertical rdbGrid[1,i-1] = r[i]
                        else rdbGrid[i-1,1] = r[i] end
                        signal_connect(callback_rdb,  r[i], "toggled", Nothing, (), false, name)
                end
                g[2,row] = rdbGrid
        elseif type == "S"      # slider
                sliderLimit = [0.0,100.0]
                re = r"^[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)$";
                if occursin(re, field[4])
                        sliderLimit[1] = parse(Float64,field[4])
                end
                if occursin(re, field[5])
                        sliderLimit[2] = parse(Float64,field[5])
                end
                if sliderLimit[1] > sliderLimit[2] 
                        pushfirst!(sliderLimit,sliderLimit[2])
                end
                slider = GtkScale(false, sliderLimit[1]:sliderLimit[2])
                set_gtk_property!(slider, :value_pos, 1)        # value at right
                delta = sliderLimit[2] - sliderLimit[1]
                digits = floor(log10(delta)-2) < 1 ? -floor(log10(delta)-2) : 0
                hasDefault(name,string(sliderLimit[1] + 0.5 * delta))
                set_gtk_property!(slider, :digits, digits)
                set_gtk_property!(slider, :name, "fg_scale")
                if delta * 10^digits < 101
                	set_gtk_property!(slider,:margin_right,40)                   	
                end
                g[2,row] = slider
                wdgHandles[name] = slider
        else	# unknown
        	g[1,row] = set_gtk_property!(g[1,row],:label,"Unknown")
        	g[2,row] = GtkLabel(join(field," "))             
        end
        global row = row + 1
end
# buttons *******************************************************
if !oneCombo
        buttonGrid = GtkGrid()
        set_gtk_property!(buttonGrid, :column_homogeneous, true)
        ok = createButton("Ok",standardButtons["Ok"])
        cancel = createButton("Cancel",standardButtons["Cancel"])
        reset = createButton("Reset",standardButtons["Reset"])
        buttonGrid[1,1] = ok
        wdgHandles["Ok"] = ok
        signal_connect((widget, event) -> callback_ok(ok),ok, "button-press-event")
        buttonGrid[2,1] = cancel
        id = signal_connect(cancel, "button-press-event") do widget, event
                button = get_gtk_property(widget,:name,String)
                destroy(fg_win)
                callBack(Dict("fg_button" => "Cancel"))
        end
        buttonGrid[3,1] = reset
        id = signal_connect(reset, "button-press-event") do widget, event
                setValues()
        end
        g[1:2,row] = buttonGrid
end
# styles ********************************************************

for (key,style) in wdgStyles
        if !haskey(wdgHandles,key)
	        println("Style ($style) on field $key that doesn't exists")
	        continue
	end
	fg_set_gtk_style(wdgHandles[key],"* {$style}")
end
# events ********************************************************
for (field) in wdgEvents
	name,event,functionCall = field[2:4]
	if lowercase(functionCall) == "call" functionCall = field[5] end
	item = split(functionCall,r"\s+",limit=2)    # my be a parameter space separated	
	event = lowercase(event)
        if lowercase(functionCall) ∉ ["submit","exit","call"]
        	functionCall = item[1]
        	if !isdefined(formgen,Symbol(functionCall)) && !isdefined(Main,Symbol(functionCall))               
                        println("Main or formgen doesn't contain $functionCall function")
                        continue
                end
        end
        if event in ["start","onstart"]	# Start event *******************
                	eval(Meta.parse("Main.$functionCall(\"Start\")"))
                	continue
        end
        if event in ["end","onend"]	# End event *******************
                	onEndFunction = functionCall
                	continue
        end
        if !haskey(wdgHandles,name)
                println("Event on field that doesn't exists: ",name)
                continue
        end       
        wdgHandle = wdgHandles[name]
        if lowercase(functionCall) == "submit"	# submit form
                wdgEventsId[field[2]] = signal_connect(wdgHandle, event) do widget
                        callback_ok(widget)
                end
                continue
        end
        extraParm = length(item) > 1 ? ",\"" * item[2] * "\"" : ""
        if event == "change" && wdgTypes[name]  == "CMB"
                wdgEventsId[name] = signal_connect(wdgHandle, "changed") do widget
                         eval(Meta.parse("Main.$functionCall(\"" * name *"\"$extraParm)"))
                end
        else
                link = getParameter(field[4:end],"link")
                link = link != "" ? ",\"$link\"$extraParm" : extraParm
                if isdefined(formgen,Symbol(functionCall))
                        signal_connect(wdgHandle, event) do widget
                                eval(Meta.parse("$functionCall(\"$name\"$link)"))
                        end                                       
                else 
                        signal_connect(wdgHandle, event) do widget
                                eval(Meta.parse("Main.$functionCall(\"$name\"$link)"))
                        end
                end

        end
end
set_gtk_property!(g, :margin, 5)
set_gtk_property!(g, :row_spacing, 10)  # introduce a 10-pixel gap between rows
push!(fg_win,g)
setValues()     # set defaults
showall(fg_win)
if !isinteractive()
    	c = Condition()
	signal_connect(fg_win, :destroy) do widget
        fg_reset_CSS(screen_provider)   # remove screen styles (window ?)
    	notify(c)
	end
	@async Gtk.gtk_main()
	wait(c)
end
end     # FormGen end
end     # module end
