# Julia Form generator **********************************
# sandbox 0.1.2 13 May 2023
# BSD Zero Clause License free to use but no warranties
# El Condor - Condor Informatique - Turin
# *******************************************************
using Dates
using Gtk, Gtk.ShortNames
using GTK3_jll 
const libgtk = libgtk3
const libgdk = libgdk3
include("formgen.jl")
global version = "0.1.3 13 May 2023"
dir = @__DIR__
#println("$dir\\quotes.txt")
global lines              # quote array
global authorQuotes = []  # Author selected quote
global Author
global nQuote = 0         # quote position
function formAnswer(data)
        println("******* Data *******")
        for key in keys(data)
            dt = (typeof(data[key]) == "Float64") ? string(data[key]) : data[key]
            println("$key\t$dt")
        end
end
function About(x)
    global version
    signal = """
    Julia Form Generator by GTK
    Version $version
    El Condor www.condorinformatique.com Turin
    """
    info_dialog(signal)
end
function changeCombo(x,y)
    formgen.createOptions("months",formgen.getDateNames("%#Days"))
end
function Examples(example)
    println(example)
    frm = "T text"
    if example == "General" frm = frmGeneral
    elseif example == "Activate" frm = frmActivate
    elseif example == "Buttons" frm = frmButtons
    elseif example == "Controls" frm = frmControls
    elseif example == "Events" frm = frmEvents
    elseif example == "Numbers" frm = frmNumbers
    elseif example == "Only combo" frm = frmOnlyCombo
    elseif example == "RdbChecks" frm = frmRdbChecks
    elseif example == "Sliders" frm = frmSliders
    elseif example == "Styles" frm = frmStyles
    elseif example == "Texts" frm = frmTexts
    end
    set_gtk_property!(frmList,:text,frm)
    formgen.formGen(frm,formAnswer)
end
function Functions(example)
    box = GtkButtonBox(:v)
    if example == "Calendar" 
        Title = "Create Calendar"
        push!(box,formgen.fg_create_calendar(receiveDate,"2023.11.16","E d U Y"))
    elseif example == "Text view" 
        Title = "Create Text view monospaced"
        handlers = formgen.fg_create_textview(width=300,height=500)
        set_gtk_property!(handlers[1],:text,"Monospaced text\nBy Condor")
        ccall((:gtk_text_view_set_monospace,libgtk),Cvoid,(Ptr{GObject},Cuchar),handlers[2],true)
        push!(box,handlers[3])   # is scroll area
    end
    win = GtkWindow("$Title")
    formgen.fg_set_gtk_style(win,"* {border:3px inset red;margin-left:5px}")
    push!(win, box)
    showall(win)
end
function receiveDate(x::String)
    info_dialog(x)
  end
function roots(btn,widget)
    local n = formgen.valueOf(widget)
    if btn == "cubicRoot"
        println(cbrt(formgen.valueOf(widget)))
        info_dialog("Cubic root of $n = " * string(cbrt(formgen.valueOf(widget))))
    else
        info_dialog("Square root of $n = " * string(sqrt(formgen.valueOf(widget))))
    end
end
function infoLang(btn,widget)
	info_dialog("Languages: " * string(formgen.valueOf("Languages")))
end
function closeForm(p)
    destroy(win)
    signal_connect(win, :destroy) do widget
 #       Gtk.gtk_quit()
    end
    println("See you later alligator.")
end
function loadCite(widget)
    empty!(authorQuotes)
    global Author = formgen.valueOf("Authors")
    for line in lines
        aline = split(line,"\t")
        if aline[1] == Author push!(authorQuotes,aline[2]) end
    end
    global nQuote = 1
    browseQuote(widget,"Authors")
end
function checkAuthor(event)
Author = formgen.valueOf("Authors")
	println(Author)
	if Author == "Alan Perlis" return true end
	info_dialog("Author mus be Alan Perlis!")
	return false
end
function loadAuthors(event)
    f = open(joinpath(dir, "quotes.txt")); 
    global lines = readlines(f)
    Authors = Dict{String,String}()
    for line in lines
        aline = split(line,"\t")
        Authors[aline[1]] = ""
    end
    aAuthors = sort!(collect(keys(Authors)))
    sAuthors = join(aAuthors,"|")
    formgen.createOptions("Authors",sAuthors)
    close(f)
    commentStyle = "* {border:3px inset;margin:5px;padding:3px;background:silver}"
    formgen.fg_set_gtk_style(formgen.wdgHandles["c"],commentStyle)
end
function browseQuote(name,button)
    global nQuote,Author
    nQ = length(authorQuotes)
    if nQ > 0
        incr = button != "Next" ? -1 : 1
        nQuote = rem(nQuote+incr+nQ,nQ)
        formgen.setValue("c",authorQuotes[nQuote+1])
        formgen.setValue("AuthorInfo","$Author quotes $nQ quote n° " * string(nQuote+1))
    end
end
function setOnStart(event)
    TextView = formgen.wdgHandles["TextView"][2]
    ccall((:gtk_text_view_set_monospace,libgtk),Cvoid,(Ptr{GObject},Cuchar),TextView,true)
    formgen.fg_set_gtk_style("Date","* {color:red}")	
    formgen.fg_set_gtk_property("fg_window",:resizable,false)
end
function actionDB(widget,after="")  
    println(formgen.valueOf(after == "" ? widget : after))
end
date = Dates.format(now(), "yyyy-mm-dd HH:MM:SS")
frmActivate = """
window 'Activate on enter or change combo' ground #C000C00
T txt Activate onActivate Submit hint 'Enter text and hit enter key'
CMB months 'Months' %Months Value June onChange Submit
"""
frmButtons = """
Window 'Buttons and static form' background #00C0C0 reset static
Menu _File Close Close
Menu _Info About About
T New '' onActivate actionDB
T Update '' onActivate actionDB width 200
T Cancel '' onActivate actionDB
B bNew Insert after New Call actionDB width 60
B bUpdate Update after Update Call actionDB
B bCancel Delete after Cancel Call actionDB
Style bUpdate 'min-width:60px'
Style bCancel 'min-width:60px'
Style bCuCu 'min-width:100px'
"""
frmControls = """
Window 'Controls' background #00C0C0
label right color blue
PN positiveNumber 'Waiting time'
Control positiveNumber <= 1000 'waiting too long'
Required positiveNumber
Text e_mail Mail
Control e_mail is mail 'Incorrect mail form'
Slider Min 'Min pressure' 30 150
Slider Max 'Max pressure' 40 220
Control Min < Max 'Minimum must be less of Maximum'
PSW password '' hint '8 characters'
Control password is '^.{8,}\$' 'Almost 8 characters'
"""
frmEvents = """
Window 'Events' background #00C0C0 onStart loadAuthors onEnd checkAuthor
Comment c 'Quote' color green height 200
Navigator browse browseQuote
COMBO Authors "" onChange loadCite
C AuthorInfo
"""
frmNumbers = """
Window 'Numbers form' background #00C0C0
N naturalNumber
PN positiveNumber 'Waiting time' value 345
FN floatingNumber
S slider006 'Slider two decimal' -3 3 value -1.2
S Slider111 'Inverted limits' 111 11
Required naturalNumber positiveNumber
Control positiveNumber <= 1000 'waiting too long'
B sqrt \u221a after positiveNumber call roots
B cubicRoot \u221b After floatingNumber call roots
Default floatingNumber -792.5"""
frmGeneral = """
Window 'Form (de)generator' ground #ADD8E6
CMB Greek 'Combo example' Α:Alfa|Β:Beta|Δ:Delta|Γ:Gamma
CMB months '' %Months Value June
B change '' after months call changeCombo
PN waitingTime
B sqrt \u221a after waitingTime call roots
S slider '' -10 10 value -3.7
C cmt 'This is a blue center aligned comment' color blue center
CHECK Agree '' 'Consent sending infos'
Text user_name
Rdb Gender '' 'M:Male|F:Female|N:Not exposed' Default F
Un unMod
U notModifiable '' default "Not modifiable field"
H Timestamp '$date'
B Ok \u2714
B Cancel \u2718
B Reset \u21B6
Default user_name 'Joe Mds' unMod 3.14159 waitingTime 99
Default Agree Off Greek Δ
"""
frmOnlyCombo = """
Window 'Only one combo' ground Olive
CMB Combo '' 'B:Buttons|G:General|N:Numbers|OC:Only combo|T:Text'
"""
frmRdbChecks = """
window 'Rdb and Checks'
Rdb Gender '' 'M:Male|F:Female|N:Not exposed'
Rdb Liking '' 'L:Low|M:Medium|H:High' vertical
CHECK Agree '' 'Consent sending infos'
CKL Languages '' 'C:C, C++, C#|JS:JavaScript|Julia|PHP|Python|Others' default Julia
Default PHP on Liking H
B Info \u2139 after Languages Call infoLang
"""
frmSliders = """
window 'Snow babaciu' ground cyan
label left
Slider Temperature '' 34 44 value 36.5
Slider Min 'Min pressure' 30 150
Slider Max 'Max pressure' 40 220
Default Max 140
Control Min < Max 'Minimum must be less of Maximum'
"""
frmStyles = """
window 'Style example' ground teal
label right after : color yellow
CMB months '' %Months
CKB Agree '' 'Consent sending infos'
Style Agree 'border:3px outset olive;background-color:silver'
Style months 'border:3px outset olive;border-radius:3px'
"""
frmTexts = """
Window 'Texts example' ground silver onStart setOnStart
label color navy after ' :'
Text Name '' hint "Name surname"
T TextView '' height 100 default 'set_gtk_property!(textBuf2,:text,"Inserted text")'
DATE Date 'Job Start' Default tomorrow title 'Action date'
PSW password '' hint '8 characters'
Control password is '^.{8}\$' '8 characters'
DATE BirthDate '' Default 2023-1-30 Format "E d U Y"
H Timestamp '$date'
H hidden 'A hidden field'
Menu Info Info 'Sandbox $version'
U notModifiable '' value 'Not modifiable field'
File FileIn '' filter '*.jpg,*.png,*.gif' title 'Take a graphic file' hint 'Image file'
Folder Folder '' 'Images folder' hint 'Images folder'
Default FileIn marka.png
"""
menu = [["File", "End", "closeForm"],
["Examples", "Activate", "Examples"],
["Examples", "Buttons", "Examples"],
["Examples", "Controls", "Examples"],
["Examples", "Events", "Examples"],
["Examples", "General", "Examples"],
["Examples", "Numbers", "Examples"],
["Examples", "Only combo", "Examples"],
["Examples", "RdbChecks", "Examples"],
["Examples", "Sliders", "Examples"],
["Examples", "Styles", "Examples"],
["Examples", "Texts", "Examples"],
["Widgets", "Calendar", "Functions"],
["Widgets", "Text view", "Functions"],
["Info", "About", "About"]]
g = GtkGrid()
set_gtk_property!(g, :row_spacing, 1) 
g[1:3,1] = formgen.fg_create_menu(menu)
handlers = formgen.fg_create_textview(width=728,height=450)
ccall((:gtk_text_view_set_monospace,libgtk),Cvoid,(Ptr{GObject},Cuchar),handlers[2],true)
frmList = handlers[1]
btn = GtkButton("Try")
id = signal_connect(btn, "button-press-event") do widget, event
    formgen.formGen(get_gtk_property(frmList, :text, String),formAnswer)
end
g[3,3] = btn
g[1:5,2] = handlers[3]
win = GtkWindow( "FormGen Sandbox")
formgen.fg_set_gtk_style(handlers[3],"* {border-width:2px;margin:5px}")
push!(win, g)
showall(win)
#Examples("Numbers")
if !isinteractive()
    c = Condition()
    signal_connect(win, :destroy) do widget
    notify(c)
end
@async Gtk.gtk_main()
wait(c)
end
